/////////////////////////////////////////////////////////
//
//  zCountSpace.cc
//
//  Implements:
//      - Cudd_zddCountSpace
//        Measures the explored search space eliminated by 
//        a set of clauses (usually conflict-induced clauses)
//
//  Notes:
//      - Used with Satometer [Aloul et al. DAC'02]
//      - Inorder to turn off the EXACT approach, comment the three
//        cuddZddFixRange functions in cuddZddCountSpace.
//
// by Fadi Aloul, University of Michigan
// (c) Copyrights Fadi Aloul, 2002.
//
/////////////////////////////////////////////////////////

#include    "zLib.h"
#include    "st.h"
#include    "cuddInt.h"
#include    "math.h" 

/* Macros. */
#define MAX(a, b)   ((a) > (b) ? (a) : (b))
#define MIN(a, b)   ((a) < (b) ? (a) : (b))

double ONE=1, ZERO=0;
double DoExactInCount = 1;

typedef char int8;


/* Fcns */

static enum st_retval st_zdd_count_dbl_free ARGS((char *key, char *value, 
						  char *arg));

// Space count
static int8 cuddZddCountSpace ARGS((DdManager *zdd, DdNode * P, 
				      st_table *tableLO, 
				      st_table *tableHI, DdNode * base, 
				      DdNode * empty, 
				      double *lo, double *hi, 
				      const bool storeInCache));

static void cuddZddFixRange ARGS((DdManager *zdd,DdNode * E,DdNode * T, 
				  st_table *tableLO, st_table *tableHI,
				  DdNode * base,DdNode * empty,double *lo,
				  double *hi));
static void cuddZddFindRange ARGS(( double *loG, double *hiG, double *loH,
				    double *hiH, double *lo, double *hi )); 

static void Int_SUB ARGS((double *loA,double *hiA,double *loB,double 
			  *hiB,double *lo,double *hi));
static void Int_ADD ARGS((double *loA,double *hiA,double *loB,double 
			  *hiB,double *lo,double *hi));
static void Int_MAX ARGS((double *loA,double *hiA,double *loB,double 
			  *hiB,double *lo,double *hi));
static void Int_MIN ARGS((double *loA,double *hiA,double *loB,double 
			  *hiB,double *lo,double *hi));





////////////////////////////
// Main CountSpace function
//
////////////////////////////

double
Cudd_zddCountSpace(
  DdManager * zdd,
  DdNode * P,
  double *lo,
  double *hi, 
  double doExact)
{
    double	res, HI, LO;
    DdNode	*base, *empty;
    st_table	*tableLO, *tableHI;
    
    base  = DD_ONE(zdd);
    empty = DD_ZERO(zdd);
    
    tableLO = st_init_table(st_ptrcmp, st_ptrhash);
    if (tableLO == NULL) return((double)CUDD_OUT_OF_MEM);
    tableHI = st_init_table(st_ptrcmp, st_ptrhash);
    if (tableHI == NULL) return((double)CUDD_OUT_OF_MEM);

    DoExactInCount = doExact;

    res = cuddZddCountSpace(zdd, P, tableLO, tableHI, base, empty, &LO, &HI, 1);
    *lo = LO;
    *hi = HI;

    st_foreach(tableLO, st_zdd_count_dbl_free, NIL(char));
    st_free_table(tableLO);    

    st_foreach(tableHI, st_zdd_count_dbl_free, NIL(char));
    st_free_table(tableHI);    

    return(res);

} 


////////////////////////////
// Interval Arithmatic
//
////////////////////////////

static void Int_SUB(
		    double *loA,
		    double *hiA,
		    double *loB,
		    double *hiB,
		    double *lo,
		    double *hi) {
  double loC, hiC;
  
  loC = *loA - *hiB;
  hiC = *hiA - *loB;
  
  if(loC<=hiC) {
    *lo = loC;
    *hi = hiC;
  } else {
    *lo = hiC;
    *hi = loC;
  }
}

static void Int_ADD(
		    double *loA,
		    double *hiA,
		    double *loB,
		    double *hiB,
		    double *lo,
		    double *hi) {
  double loC, hiC;
  
  loC = *loA + *loB;
  hiC = *hiA + *hiB;
  
  if(loC<=hiC) {
    *lo = loC;
    *hi = hiC;
  } else {
    *lo = hiC;
    *hi = loC;
  }
}

static void Int_MIN(
		    double *loA,
		    double *hiA,
		    double *loB,
		    double *hiB,
		    double *lo,
		    double *hi) {
  double loC, hiC;
  
  loC = MIN(*loA, *loB);
  hiC = MIN(*hiA, *hiB);
  
  if(loC<=hiC) {
    *lo = loC;
    *hi = hiC;
  } else {
    *lo = hiC;
    *hi = loC;
  }
}

static void Int_MAX(
		    double *loA,
		    double *hiA,
		    double *loB,
		    double *hiB,
		    double *lo,
		    double *hi) {
  double loC, hiC;
  
  loC = MAX(*loA, *loB);
  hiC = MAX(*hiA, *hiB);
  
  if(loC<=hiC) {
    *lo = loC;
    *hi = hiC;
  } else {
    *lo = hiC;
    *hi = loC;
  }
}



////////////////////////////
// Approximate Computation
//
// Identifies the range of the product
//
////////////////////////////

static void 
cuddZddFindRange(
		 double *loG,
		 double *hiG,
		 double *loH,
		 double *hiH,
		 double *lo,
		 double *hi
		 ) 
{

  double      LOa,LOb,LOc,LOd, HIa,HIb,HIc,HId;

  // Range:  [max(0, |g|+|h|-1), min(|g|,|h|)]

  // min(g,h)
  Int_MIN(loG,hiG,loH,hiH, &LOa, &HIa);
  
  // g+h
  Int_ADD(loG,hiG,loH,hiH, &LOb, &HIb);

  // g+h - 1
  Int_SUB(&LOb,&HIb,&ONE,&ONE, &LOc, &HIc);
  
  // max(0, g+h-1)
  Int_MAX(&ZERO,&ZERO,&LOc,&HIc, &LOd, &HId);
  
  // [max,min]...
  if(LOd <= LOa)
    *lo = LOd;
  else
    *lo = LOa;
  
  if(HId >= HIa)
    *hi = HId;
  else            
    *hi = HIa; 
  
  return;
}


////////////////////////////
// Controlled-Accuracy Computation
//
// Minimizes the range GAP according to 
// a user-specified limit
//
////////////////////////////

static void
cuddZddFixRange(
		DdManager *zdd,
		DdNode * E,
		DdNode * T,
		st_table * tableLO,
		st_table * tableHI,
		DdNode * base,
		DdNode * empty,
		double *lo,
		double *hi)    
{

  // If equal do nothing (dont change values, just return)
  if(*lo == *hi) 
    return;
  
  double LOa, HIa;

  // Union(E,T) -> CountSpace of the result
  DdNode* tmp= Cudd_zddUnionS(zdd, E, T);   // Can use SA for reduction **
  Cudd_Ref(tmp);

  // Make sure you dont store in Cache.. this node will be deleted next!
  cuddZddCountSpace(zdd, tmp, tableLO, tableHI, base, empty, &LOa, &HIa, 0);
  

  // Sanity check
  if(LOa != HIa && DoExactInCount == 0) {
    printf("Hmmm.... error in cuddZddFixRange %f != %f\n", LOa, HIa);
  }
  
  // Ok done with this BDD node.. free it!
  Cudd_RecursiveDerefZdd(zdd, tmp);

  // Update values
  *lo = LOa;
  *hi = HIa;

  return;

}


////////////////////////////
// Measure Search Space Coverage!
//
////////////////////////////

static int8
cuddZddCountSpace(
  DdManager *zdd,
  DdNode * P,
  st_table * tableLO,
  st_table * tableHI,
  DdNode * base,
  DdNode * empty,
  double *lo,
  double *hi, 
  const bool storeInCache)     
{

    if (P == empty) {
      *lo = 1;
      *hi = 1;
      return(0);

    }
    if (P == base) {
      *lo = 0;
      *hi = 0;
      return(1);
    }


    // Lookup CACHE first
    double	*dummy, *dummy1;
    if ( st_lookup(tableLO, (char *)P, (char **)(&dummy)) &&
	 st_lookup(tableHI, (char *)P, (char **)(&dummy1)) ) {
      *lo = *dummy;
      *hi = *dummy1;
      return(2);
    }


    // Not in Cache...
    double      loT=0, hiT=0;
    double      LOf, HIf;
        
    // *******  TWO Possible cases... xT=h, xE=g, gT=t, gE=q *******
    // 1) X's E edge NOT pointing to -X
    //    ||f|| = 0.5 [ ||g|| + ||g.h|| ]

    // 2) X's E edge IS pointing to -X
    //    ||f|| = 0.5 [ ||t.q|| + ||q.h|| ]
    
    // Find ||h||
    cuddZddCountSpace(zdd, cuddT(P), tableLO, tableHI, base, empty, &loT, &hiT, 
		      storeInCache);
    

    if( (cuddE(P)->index & 1) && 
	(cuddE(P)->index == (P->index + 1))) {
      // IF(xE == -X) find ||q|| & ||t||
      double      loET=0, hiET=0, loEE=0, hiEE=0;
      double      LOa, HIa, LOb, HIb;
      cuddZddCountSpace(zdd, cuddE(cuddE(P)), tableLO, tableHI, base, 
				empty, &loEE, &hiEE, storeInCache);
      cuddZddCountSpace(zdd, cuddT(cuddE(P)), tableLO, tableHI, base, 
				empty, &loET, &hiET, storeInCache);

      // ||q.h||
      cuddZddFindRange(&loEE, &hiEE, &loT, &hiT, &LOa, &HIa);
      if((HIa - LOa) > DoExactInCount) {
	cuddZddFixRange(zdd, cuddE(cuddE(P)), cuddT(P), tableLO, tableHI,
			base, empty, &LOa, &HIa);
      }

      // ||t.q||
      cuddZddFindRange(&loET, &hiET, &loEE, &hiEE, &LOb, &HIb);
      if((HIb - LOb) > DoExactInCount) {
	cuddZddFixRange(zdd, cuddE(cuddE(P)), cuddT(cuddE(P)), tableLO, tableHI,
			base, empty, &LOb, &HIb);
      }

      // ||t.q|| + ||q.h||
      Int_ADD(&LOa,&HIa,&LOb,&HIb, &LOf, &HIf);

    } else {
      // IF(xE != -X) find ||g||
      double      loE=0, hiE=0;
      double      LOe, HIe;
      cuddZddCountSpace(zdd, cuddE(P), tableLO, tableHI, base, 
			       empty, &loE, &hiE, storeInCache);

      //find ||g.h||
      cuddZddFindRange(&loE, &hiE, &loT, &hiT, &LOe, &HIe);
      if((HIe - LOe) > DoExactInCount) {
	cuddZddFixRange(zdd, cuddE(P), cuddT(P), tableLO, tableHI, 
			base, empty, &LOe, &HIe);
      }

      // ||g|| + ||g.h||   
      Int_ADD(&loE,&hiE,&LOe,&HIe, &LOf, &HIf);
    }

    
    // DONE.... final computation

    // 0.5 * ..
    *lo = 0.5*LOf;
    *hi = 0.5*HIf;


    // Lets store the result in Cache
    if(storeInCache) {
      // LO Table
      dummy = ALLOC(double, 1);
      if (dummy == NULL) {
	return(CUDD_OUT_OF_MEM);
      }
      *dummy = *lo;
      if (st_insert(tableLO, (char *)P, (char *)dummy) == ST_OUT_OF_MEM) {
	FREE(dummy);
      return(CUDD_OUT_OF_MEM);
      }
      
      // HI Table
      dummy1 = ALLOC(double, 1);
      if (dummy1 == NULL) {
	return(CUDD_OUT_OF_MEM);
      }
      *dummy1 = *hi;
      if (st_insert(tableHI, (char *)P, (char *)dummy1) == ST_OUT_OF_MEM) {
	FREE(dummy1);
	return(CUDD_OUT_OF_MEM);
      }
    } // if(storeInCache)


    return(2);

}


static enum st_retval
st_zdd_count_dbl_free(
  char * key,
  char * value,
  char * arg)
{
    double	*d;

    d = (double *)value;
    FREE(d);
    return(ST_CONTINUE);

}
