/////////////////////////////////////////////////////////
//
//  zSubsume.cc
//
//  Implements:
//      - Cudd_zddNoSub
//        Eliminates Subsumed clauses and tautologies
//      - Cudd_zddNoSubA 
//        Eliminates subsumed clauses and tautologies
//        and performs a combination of Boolean Algebraic 
//        manipulations (i.e. absorbtion, subsumption,
//        and resolution).
//
//  Notes:
//      - Cudd_zddNoSub was proposed in [Chatalic and Simon, ICTAI'00]  
//      - Cudd_zddNoSubA was proposed in [Aloul et al., IWLS'01]  
//
// by Fadi Aloul, University of Michigan
// (c) Copyrights Fadi Aloul, 2002.
//
/////////////////////////////////////////////////////////


#include    "zLib.h"


// Subsumption only
DdNode	*
Cudd_zddNoSub(
  DdManager * dd,
  DdNode * f)
{
    DdNode	*res;

    do {
	dd->reordered = 0;
	res = cuddZddNoSub(dd, f);
    } while (dd->reordered == 1);
    return(res);

} /* end of Cudd_zddNoSub */


DdNode	*
Cudd_zddNoSubA(
  DdManager * dd,
  DdNode * f)
{
    DdNode	*res;

    do {
      dd->reordered = 0;
      res = cuddZddNoSubA(dd, f);
    } while (dd->reordered == 1);
    
    return(res);

} /* end of Cudd_zddNoSub */


/*****************************************************/

/*****************************
 * SubDiff
 * -optimized-
 ****************************/

DdNode* 
cuddZddSubDiff(
  DdManager * dd,
  DdNode * f,
  DdNode * g )
{
  DdNode	*one = DD_ONE(dd);
  DdNode	*zero = DD_ZERO(dd); 

  statLine(dd);

  if(f == zero ) {
    return (zero);
  }

  if(f == one && g == one) {
    return (zero);
  }

  if(f == one) {
    return (one);
  }

  if(g == zero) {
    return (f);
  }

  if(g == one) {
    return (zero);
  }
  

  // Cache Lookup
  DdNode *tmp = cuddCacheLookup2Zdd(dd, cuddZddSubDiff, f, g);
  if (tmp)
    return(tmp);

  DdNode        *tmp0, *tmp1, *tmp2;
  int           f_index, g_index;

  f_index = cuddIZ(dd,f->index);
  g_index = cuddIZ(dd,g->index);
 
  if(f_index > g_index) {
    
    tmp = cuddZddSubDiff(dd, f, cuddE(g));
    if (tmp == NULL) {
      return(NULL);
    }
    //cuddRef(tmp);

  }
  else if(f_index < g_index) {

    tmp1 = cuddZddSubDiff(dd, cuddT(f), g);
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);

    tmp0 = cuddZddSubDiff(dd, cuddE(f), g);
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);
    
    tmp = cuddZddGetNode(dd, f->index, tmp1, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    //cuddRef(tmp);
    cuddDeref(tmp1); // Cudd_RecursiveDerefZdd(dd, tmp1);
    cuddDeref(tmp0); // Cudd_RecursiveDerefZdd(dd, tmp0);
    
  }
  else {

    tmp1 = cuddZddSubDiff(dd, cuddT(f), cuddT(g));
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);
    
    tmp2 = cuddZddSubDiff(dd, tmp1, cuddE(g));
    if (tmp2 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp2);
    Cudd_RecursiveDerefZdd(dd, tmp1);

    tmp0 = cuddZddSubDiff(dd, cuddE(f), cuddE(g));
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp2);
      return(NULL);
    }
    cuddRef(tmp0);

    tmp = cuddZddGetNode(dd, f->index, tmp2, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp0);
      Cudd_RecursiveDerefZdd(dd, tmp2);
      return(NULL);
    }
    // cuddRef(tmp);
    cuddDeref(tmp2); //Cudd_RecursiveDerefZdd(dd, tmp2);
    cuddDeref(tmp0); //Cudd_RecursiveDerefZdd(dd, tmp0);
    
  }
  
  cuddCacheInsert2(dd, cuddZddSubDiff, f, g, tmp);
  //Cudd_Deref(tmp);

  return tmp;  
}

/*****************************
 * Subsumption only
 * -optimized-
 ****************************/

DdNode* 
cuddZddNoSub(
  DdManager * dd,
  DdNode * f )
{
  DdNode	*one = DD_ONE(dd);
  DdNode	*zero = DD_ZERO(dd); 
  
  statLine(dd);

  if(f == zero ) {
    return (zero);
  }

  if(f == one ) {
    return (one);
  }


  DdNode *tmp = cuddCacheLookup1Zdd(dd, cuddZddNoSub, f);
  if(tmp)
    return (tmp);

  DdNode        *tmp0, *tmp1, *tmp2, *childE, *childT;
  int           f_index;

  // Should this point to permZ ?? -- no
  f_index = f->index;

  childT = cuddT(f);
  childE = cuddE(f);  


  // ORDER is IMPORTANT!!

  // 1) Pointing to same node? DELETE! (a+b)(b) => (b)
  if(cuddT(f) == cuddE(f)) {
    tmp = cuddZddNoSub(dd, cuddT(f));
    if (tmp == NULL) {
      return(NULL);
    }
    //cuddRef(tmp);
    
  } else 
    // 3) Tautology check.. (SHOULD BE before last..) if(d+!d+..) = 1;
    if((childT->index == f_index + 1) && 
       (childT->index & 1)) {
      
      tmp1 = cuddZddNoSub(dd, cuddE(childT));
      if (tmp1 == NULL) {
	return(NULL);
      }
      cuddRef(tmp1);
      
      tmp0 = cuddZddNoSub(dd, cuddE(f));
      if (tmp0 == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp1);
	return(NULL);
      }
      cuddRef(tmp0);
      
      tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
      if (tmp2 == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp1);
	return(NULL);
      }
      cuddRef(tmp2);
      Cudd_RecursiveDerefZdd(dd, tmp1);
      
      tmp = cuddZddGetNode(dd, f_index, tmp2, tmp0);
      if (tmp == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp2);
	return(NULL);
      }
      //cuddRef(tmp);
      cuddDeref(tmp2); //Cudd_RecursiveDerefZdd(dd, tmp2);
      cuddDeref(tmp0); //Cudd_RecursiveDerefZdd(dd, tmp0);
      
    } else 
      
      // 4) Regular div subsume...
      {
	tmp1 = cuddZddNoSub(dd, cuddT(f));
	if (tmp1 == NULL) {
	  return(NULL);
	}
	cuddRef(tmp1);
		
	tmp0 = cuddZddNoSub(dd, cuddE(f));
	if (tmp0 == NULL) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  return(NULL);
	}
	cuddRef(tmp0);
	
	tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
	if (tmp2 == NULL) {
	  Cudd_RecursiveDerefZdd(dd, tmp0);
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  return(NULL);
	}
	cuddRef(tmp2);
	Cudd_RecursiveDerefZdd(dd, tmp1);
	
	tmp = cuddZddGetNode(dd, f_index, tmp2, tmp0);
	if (tmp == NULL) {
	  Cudd_RecursiveDerefZdd(dd, tmp0);
	  Cudd_RecursiveDerefZdd(dd, tmp2);
	  return(NULL);
	}
	//cuddRef(tmp);
	cuddDeref(tmp2); //Cudd_RecursiveDerefZdd(dd, tmp2);
	cuddDeref(tmp0); //Cudd_RecursiveDerefZdd(dd, tmp0);
	
      }
  
  cuddCacheInsert1(dd, cuddZddNoSub, f, tmp);
  //Cudd_Deref(tmp);
  
  return tmp;    
}


/*****************************
 * Subsumption & Algebraic Manipulation Techniques
 * -opt-
 ****************************/

DdNode* 
cuddZddNoSubA(
  DdManager * dd,
  DdNode * f )
{
  DdNode	*one = DD_ONE(dd);
  DdNode	*zero = DD_ZERO(dd); 
  
  statLine(dd);

  if(f == zero ) {
    return (zero);
  }

  if(f == one ) {
    return (one);
  }


  DdNode *tmp = cuddCacheLookup1Zdd(dd, cuddZddNoSubA, f);
  if(tmp)
    return (tmp);

  DdNode        *tmp0, *tmp1, *tmp2, *childE, *childT;
  int           f_index;
  bool         RegRed = 1, NoRed = 1;

  f_index = f->index;

  childT = cuddT(f);
  childE = cuddE(f);  

  // ORDER is IMPORTANT!!

  // 1) Pointing to same node? DELETE! (a+b)(b) => (b)
  if(cuddT(f) == cuddE(f)) {
    RegRed = 0;
    tmp = cuddZddNoSubA(dd, cuddT(f));
    if (tmp == NULL) {
      return(NULL);
    }
    cuddRef(tmp);


    // CASE (A) If x followed by !x
  } else if( (childE->index == (f_index+1)) &&
	     (childE->index & 1) ) {
    RegRed = 0;


    if( cuddT(f) == one && 
	cuddT(childE) == one ) {
      // 2) If(d)(-d) -> ret 1;    

      cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
      return (one); 
      

    } else if( cuddT(f) == one ) {
      // 3) xT=1, xE=-x, -xT=z, -xE=q 
      tmp0 = cuddZddUnionSA(dd, cuddT(childE), cuddE(childE));
      if (tmp0 == NULL) {
	return(NULL);
      }
      cuddRef(tmp0);

      // Apply Difference
      if(tmp0 == one) {
	Cudd_RecursiveDerefZdd(dd, tmp0);

	cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	return (one);
      }

      // Create Node
      tmp = cuddZddGetNode(dd, f_index, one, tmp0);
      if (tmp == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	return(NULL);
      }
      cuddDeref(tmp0);


    } else if( cuddT(childE) == one ) {
      // 4)  xT=z, xE=-x, -xT=1, -xE=q => 
      // -xT=1, -xE=(zUq).. if z=-x1: -xE=(-x1EUq)            

      if(childT->index == (f_index+1)) {  // Include Tautology check...
	tmp0 = cuddZddUnionSA(dd, cuddE(childT), cuddE(childE));
      } else {
	tmp0 = cuddZddUnionSA(dd, childT, cuddE(childE));
      }
      if (tmp0 == NULL) {
	return(NULL);
      }
      cuddRef(tmp0);

      // Apply Difference
      if(tmp0 == one) {
	Cudd_RecursiveDerefZdd(dd, tmp0);

	cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	return (one);
      }

      tmp = cuddZddGetNode(dd, childE->index, one, tmp0); 
      if (tmp == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	return(NULL);
      }
      cuddDeref(tmp0);


    } else if( cuddT(f) == cuddT(childE) ) {
      // 5) x..!x..q  x__z !x__z =? zUq
      
      tmp = cuddZddUnionSA(dd, cuddT(f), cuddE(childE));
      if (tmp == NULL) {
	return(NULL);
      }
      
    } else {
      RegRed = 1;
    }

    
    // CASE (B)
  } else if( cuddT(childT) == one &&
	     cuddT(childE) == one &&
	     (
	      ((childE->index & 1) &&
	       ((childT->index +1) == childE->index)) ||
	      
	      ((childT->index & 1) &&
	       ((childE->index +1) == childT->index)) 
	      )
	     ) {
    
    RegRed = 0;

    tmp0 = cuddZddNoSubA(dd, cuddE(f));
    if (tmp0 == NULL) {
      return(NULL);
    }
    cuddRef(tmp0);
    
    // Apply Difference
    if(tmp0 == one) {
      Cudd_RecursiveDerefZdd(dd, tmp0);
      
      cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
      return (one);
    }
    
    tmp = cuddZddGetNode(dd, f_index, one, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddDeref(tmp0);
  } 

  // 3) Tautology check.. (SHOULD BE before last..) if(d+!d+..) = 1;    
  if( RegRed &&  
      (childT->index == (f_index + 1)) && 
      (childT->index & 1)) {
    
    tmp1 = cuddZddNoSubA(dd, cuddE(childT));
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);
    
    tmp0 = cuddZddNoSubA(dd, cuddE(f));
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);
    
    tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
    if (tmp2 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp0);
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp2);
    Cudd_RecursiveDerefZdd(dd, tmp1);
    
    // ===> Final Check Needed
    // Could lead to Recursive Loops.
    if( (f_index == (tmp0->index -1)) &&
	(tmp0->index & 1) ) {
      
      NoRed = 0;

      // Case 1
      if( tmp2 == one && 
	  cuddT(tmp0) == one ) {
	
	Cudd_RecursiveDerefZdd(dd, tmp2);
	Cudd_RecursiveDerefZdd(dd, tmp0);

	cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	return (one); 
		
	// Case 2
      } else if ( tmp2 == one ) {
	
	tmp1 = cuddZddUnionSA(dd, cuddT(tmp0), cuddE(tmp0));
	if (tmp1 == NULL) {
	  return(NULL);
	}
	cuddRef(tmp1);
	Cudd_RecursiveDerefZdd(dd, tmp2);
	Cudd_RecursiveDerefZdd(dd, tmp0);

	// Apply Difference
	if(tmp1 == one) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  
	  cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	  return (one);
	}
	
	// Create Node
	tmp = cuddZddGetNode(dd, f_index, one, tmp1);
	if (tmp == NULL) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  return(NULL);
	}
	cuddDeref(tmp1);	
	
	// Case 3
      } else if ( cuddT(tmp0) == one ) { 
	
	if(tmp2->index == (f_index+1)) {  // Include Tautology check...
	  tmp1 = cuddZddUnionSA(dd, cuddE(tmp2), cuddE(tmp0));
	} else {
	  tmp1 = cuddZddUnionSA(dd, tmp2, cuddE(tmp0));
	}
	if (tmp1 == NULL) {
	  return(NULL);
	}
	cuddRef(tmp1);
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp2);

	// Apply Difference
	if(tmp1 == one) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  
	  cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	  return (one);
	}
	
	tmp = cuddZddGetNode(dd, tmp0->index, one, tmp1); 
	if (tmp == NULL) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  return(NULL);
	}
	cuddDeref(tmp1);

	// Case 4
      } else if( cuddT(tmp0) == tmp2 ) { 
	
	tmp = cuddZddUnionSA(dd, tmp2, cuddE(tmp0));
	if (tmp == NULL) {
	  return(NULL);
	}
	cuddRef(tmp);
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp2);

	
	cuddCacheInsert1(dd, cuddZddNoSubA, f, tmp);
	Cudd_Deref(tmp);
	
	return tmp;  

	// No More cases.. GetNode
      } else {
	NoRed = 1;
      }
    }
    
    if(NoRed) {
      tmp = cuddZddGetNode(dd, f_index, tmp2, tmp0);
      if (tmp == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp2);
	return(NULL);
      }
      cuddDeref(tmp2);
      cuddDeref(tmp0);
    }
    
    // 4) Regular div subsume...    
  } else if(RegRed) {
    
    tmp1 = cuddZddNoSubA(dd, cuddT(f));
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);
    
    tmp0 = cuddZddNoSubA(dd, cuddE(f));
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);
    
    tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
    if (tmp2 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp0);
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp2);
    Cudd_RecursiveDerefZdd(dd, tmp1);

    // ===> Final Check Needed
    // Could lead to Recursive Loops.
    if( (f_index == (tmp0->index -1)) &&
	(tmp0->index & 1) ) {
      
      NoRed = 0;
      
      // Case 1
      if( tmp2 == one && 
	  cuddT(tmp0) == one ) {
	
	Cudd_RecursiveDerefZdd(dd, tmp2);
	Cudd_RecursiveDerefZdd(dd, tmp0);
	
	cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	return (one); 
		
	// Case 2
      } else if ( tmp2 == one ) {
	
	tmp1 = cuddZddUnionSA(dd, cuddT(tmp0), cuddE(tmp0));
	if (tmp1 == NULL) {
	  return(NULL);
	}
	cuddRef(tmp1);
	Cudd_RecursiveDerefZdd(dd, tmp2);
	Cudd_RecursiveDerefZdd(dd, tmp0);

	// Apply Difference
	if(tmp1 == one) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  
	  cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	  return (one);
	}
	
	// Create Node
	tmp = cuddZddGetNode(dd, f_index, one, tmp1);
	if (tmp == NULL) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  return(NULL);
	}
	cuddDeref(tmp1);
	
	
	// Case 3
      } else if ( cuddT(tmp0) == one ) {
	
	if(tmp2->index == (f_index+1)) {  // Include Tautology check...
	  tmp1 = cuddZddUnionSA(dd, cuddE(tmp2), cuddE(tmp0));
	} else {
	  tmp1 = cuddZddUnionSA(dd, tmp2, cuddE(tmp0));
	}
	if (tmp1 == NULL) {
	  return(NULL);
	}
	cuddRef(tmp1);
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp2);

	// Apply Difference
	if(tmp1 == one) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  
	  cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	  return (one);
	}
	
	tmp = cuddZddGetNode(dd, tmp0->index, one, tmp1); 
	if (tmp == NULL) {
	  Cudd_RecursiveDerefZdd(dd, tmp1);
	  return(NULL);
	}
	cuddDeref(tmp1); 

	// Case 4
      } else if( cuddT(tmp0) == tmp2 ) {
	
	tmp = cuddZddUnionSA(dd, tmp2, cuddE(tmp0));
	if (tmp == NULL) {
	  return(NULL);
	}
	cuddRef(tmp);
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp2);

	cuddCacheInsert1(dd, cuddZddNoSubA, f, tmp);
	Cudd_Deref(tmp);
	
	return tmp;  

	// No More cases.. GetNode
      } else {
	NoRed = 1;
      }
    }
    
    if(NoRed) {
      tmp = cuddZddGetNode(dd, f_index, tmp2, tmp0);
      if (tmp == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	Cudd_RecursiveDerefZdd(dd, tmp2);
	return(NULL);
      }
      cuddDeref(tmp2);
      cuddDeref(tmp0);
    }
  
  }
  
  cuddCacheInsert1(dd, cuddZddNoSubA, f, tmp);
  
  return tmp;    
}


/*****************************
 * Special noSub call. Used with UnionSA.
 * Doesnt involve recursive calls.
 * Only applies Algebraic manipulations 
 * at current level
 * -opt-
 ****************************/

DdNode* 
cuddZddNoSubAF(
  DdManager * dd,
  DdNode * f )
{
  DdNode	*one = DD_ONE(dd);
  DdNode	*zero = DD_ZERO(dd); 
  
  statLine(dd);

  if(f == zero ) {
    return (zero);
  }

  if(f == one ) {
    return (one);
  }


  DdNode *tmp = cuddCacheLookup1Zdd(dd, cuddZddNoSubAF, f);
  if(tmp)
    return (tmp);

  DdNode        *tmp0, *childE, *childT;
  int           f_index;


  f_index = f->index;

  childT = cuddT(f);
  childE = cuddE(f);  

  // ORDER is IMPORTANT!!

  // 1) Pointing to same node? DELETE! (a+b)(b) => (b)
  if(cuddT(f) == cuddE(f)) {
 
    cuddCacheInsert1(dd, cuddZddNoSubAF, f, cuddT(f));
    return (cuddT(f));
    

    
  } else if( (childE->index == (f_index+1)) &&
	     (childE->index & 1) ) {
    // This part checks for nodes whose 0-child has complement var

    // 2) If(d)(-d) -> ret 1;    
    if( cuddT(f) == one && 
	cuddT(childE) == one ) {
      
      cuddCacheInsert1(dd, cuddZddNoSubAF, f, one);
      return (one); 
      
      // 3) xT=1, xE=-x, -xT=z, -xE=q      
    } else if( cuddT(f) == one ) {
      
      tmp0 = cuddZddUnionSA(dd, cuddT(childE), cuddE(childE));
      if (tmp0 == NULL) {
	return(NULL);
      }
      cuddRef(tmp0);
      
      // Apply Difference
      if(tmp0 == one) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	
	cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	return (one);
      }
      
      // Create Node
      tmp = cuddZddGetNode(dd, f_index, one, tmp0);
      if (tmp == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	return(NULL);
      }
      //cuddRef(tmp);
      cuddDeref(tmp0);
      
      
      // 4)  xT=z, xE=-x, -xT=1, -xE=q => -xT=1, -xE=(zUq).. 
      //     if z=-x1: -xE=(-x1EUq)
    } else if( cuddT(childE) == one ) {
      
      if(childT->index == (f_index+1)) {  // Include Tautology check...
	tmp0 = cuddZddUnionSA(dd, cuddE(childT), cuddE(childE));
      } else {
	tmp0 = cuddZddUnionSA(dd, childT, cuddE(childE));
      }
      if (tmp0 == NULL) {
	return(NULL);
      }
      cuddRef(tmp0);
      
      // Apply Difference
      if(tmp0 == one) {
	Cudd_RecursiveDerefZdd(dd, tmp0);

	cuddCacheInsert1(dd, cuddZddNoSubA, f, one);
	return (one);
      }

      tmp = cuddZddGetNode(dd, childE->index, one, tmp0); 
      if (tmp == NULL) {
	Cudd_RecursiveDerefZdd(dd, tmp0);
	return(NULL);
      }
      //cuddRef(tmp);
      cuddDeref(tmp0); //Cudd_RecursiveDerefZdd(dd, tmp0);

      // 5) x..!x..q  x__z !x__z =? zUq
    } else if( cuddT(f) == cuddT(childE) ) {
      
      tmp = cuddZddUnionSA(dd, cuddT(f), cuddE(childE));
      if (tmp == NULL) {
	return(NULL);
      }
      //cuddRef(tmp);

    } else {
      cuddCacheInsert1(dd, cuddZddNoSubAF, f, f);
      return(f);
    }
  } else {
    cuddCacheInsert1(dd, cuddZddNoSubAF, f, f);
    return (f);
      
  }
    
  cuddCacheInsert1(dd, cuddZddNoSubAF, f, tmp);
  //Cudd_Deref(tmp);
  
  return tmp;    
}


