
/////////////////////////////////////////////////////////
//
//  zUnionS.cc
//
//  Implements:
//      - Cudd_zddUnionS
//        Unions (ANDs) Clauses with subsumption check enabled
//      - Cudd_zddUnionSA
//        Unions (ANDs) Clauses with subsumption check 
//        + algebraic manipulation techniques enabled
//      - Cudd_zddUnionReg
//        Unions (ANDs) Clauses without subsumption check
//
//  Notes:
//      - Cudd_zddUnionS was proposed in [Chatalic and Simon, ICTAI'00]  
//
// by Fadi Aloul, University of Michigan
// (c) Copyrights Fadi Aloul, 2002.
//
/////////////////////////////////////////////////////////


#include    "zLib.h"


//////////////////
// Union w/ Subsumption + 
// Algebraic manipulation techniques
//
//////////////////

DdNode	*
Cudd_zddUnionSA(
  DdManager * dd,
  DdNode * f,
  DdNode * g)
{
    DdNode	*res;

    do {
	dd->reordered = 0;
	res = cuddZddUnionSA(dd, f, g);
    } while (dd->reordered == 1);
    return(res);

} /* end of Cudd_zddUnionSA */


//////////////////
// Union w/ Subsumption 
//
//////////////////

DdNode	*
Cudd_zddUnionS(
  DdManager * dd,
  DdNode * f,
  DdNode * g)
{
    DdNode	*res;

    do {
	dd->reordered = 0;
	res = cuddZddUnionS(dd, f, g);
    } while (dd->reordered == 1);
    return(res);

} /* end of Cudd_zddUnionS */


//////////////////
// Union w/o subsumption
//
//////////////////

DdNode	*
Cudd_zddUnionReg(
  DdManager * dd,
  DdNode * f,
  DdNode * g)
{
    DdNode	*res;

    do {
	dd->reordered = 0;
	res = cuddZddUnionReg(dd, f, g);
    } while (dd->reordered == 1);
    return(res);

} /* end of Cudd_zddUnionReg */


/*****************************************************/

/*****************************
 * Union w/ Subsumption 
 * -optimized-
 ****************************/

DdNode* 
cuddZddUnionS(
  DdManager * dd,
  DdNode * f,
  DdNode * g )
{
  DdNode	*one = DD_ONE(dd);
  DdNode	*zero = DD_ZERO(dd); 

  statLine(dd);

  if(f == one || g == one ) {
    return (one);
  }

  // Cache Lookup
  DdNode *tmp = cuddCacheLookup2Zdd(dd, cuddZddUnionS, f, g);
  if (tmp)
    return(tmp);

  if(f == zero ) {
    tmp = cuddZddNoSub(dd, g);
    if (tmp == NULL) {
      return(NULL);
    }

    cuddCacheInsert2(dd, cuddZddUnionS, f, g, tmp);
    return tmp;
  }

  if(g == zero ) {
    tmp = cuddZddNoSub(dd, f);
    if (tmp == NULL) {
      return(NULL);
    }

    cuddCacheInsert2(dd, cuddZddUnionS, f, g, tmp);
    return tmp;
  }


  DdNode        *tmp0, *tmp1, *tmp2;
  int           f_index, g_index;

  f_index = cuddIZ(dd, f->index);
  g_index = cuddIZ(dd, g->index);


  if (f_index > g_index)
    return(cuddZddUnionS(dd, g, f));  

  
  if(f_index < g_index) {
    tmp1 = cuddZddNoSub(dd, cuddT(f));
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);
    
    tmp0 = cuddZddUnionS(dd, cuddE(f), g);
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);

    tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
    if (tmp2 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddRef(tmp2);
    Cudd_RecursiveDerefZdd(dd, tmp1);

    tmp = cuddZddGetNode(dd, f->index, tmp2, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp2);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }

    cuddDeref(tmp2); 
    cuddDeref(tmp0); 
    
  }
  else {
    tmp1 = cuddZddUnionS(dd, cuddT(f), cuddT(g));
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);

    tmp0 = cuddZddUnionS(dd, cuddE(f), cuddE(g));
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);
    
    tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
    if (tmp2 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddRef(tmp2);
    Cudd_RecursiveDerefZdd(dd, tmp1);

    tmp = cuddZddGetNode(dd, f->index, tmp2, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp2);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }

    cuddDeref(tmp2); 
    cuddDeref(tmp0); 

  }

  cuddCacheInsert2(dd, cuddZddUnionS, f, g, tmp);


  return tmp; 
  
}


/*****************************
 * Regular Union - no subsumption
 * -optimized-
 ****************************/

DdNode* 
cuddZddUnionReg(
  DdManager * dd,
  DdNode * f,
  DdNode * g )
{
  DdNode	*one = DD_ONE(dd);
  DdNode	*zero = DD_ZERO(dd); 

  statLine(dd);

  if(f == one || g == one ) {
    return (one);
  }

  if(f == zero ) {
    return g;
  }

  if(g == zero ) {
    return f;
  }
  
  if(f == g) {
    return f;
  }

  // Cache Lookup
  DdNode *tmp = cuddCacheLookup2Zdd(dd, cuddZddUnionReg, f, g);
  if (tmp)
    return(tmp);

  DdNode        *tmp0, *tmp1;
  int           f_index, g_index;

  f_index = cuddIZ(dd, f->index);
  g_index = cuddIZ(dd, g->index);


  if (f_index > g_index)
    return(cuddZddUnionReg(dd, g, f));  
  

  if(f_index < g_index) {
    
    tmp0 = cuddZddUnionReg(dd, cuddE(f), g);
    if (tmp0 == NULL) {
      return(NULL);
    }
    cuddRef(tmp0);

    tmp = cuddZddGetNode(dd, f->index, cuddT(f), tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }

    cuddDeref(tmp0); 
    
  }
  else {
    tmp1 = cuddZddUnionReg(dd, cuddT(f), cuddT(g));
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);

    tmp0 = cuddZddUnionReg(dd, cuddE(f), cuddE(g));
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);
    
    tmp = cuddZddGetNode(dd, f->index, tmp1, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddDeref(tmp1); 
    cuddDeref(tmp0); 

  }

  cuddCacheInsert2(dd, cuddZddUnionReg, f, g, tmp);

  return tmp; 
  
}

/*****************************
 * Union w/ Subsumption & Reduction
 * -opt-
 ****************************/

DdNode* 
cuddZddUnionSA(
  DdManager * dd,
  DdNode * f,
  DdNode * g )
{
  DdNode	*one = DD_ONE(dd);
  DdNode	*zero = DD_ZERO(dd); 

  statLine(dd);

  if(f == one || g == one ) {
    return (one);
  }

  // Cache Lookup
  DdNode *tmp = cuddCacheLookup2Zdd(dd, cuddZddUnionSA, f, g);
  if (tmp)
    return(tmp);

  if(f == zero ) {
    tmp = cuddZddNoSubA(dd, g);

    if (tmp == NULL) {
      return(NULL);
    }

    cuddCacheInsert2(dd, cuddZddUnionSA, f, g, tmp);
    return tmp;
  }

  if(g == zero ) {
    tmp = cuddZddNoSubA(dd, f);
    if (tmp == NULL) {
      return(NULL);
    }

    cuddCacheInsert2(dd, cuddZddUnionSA, f, g, tmp);
    return tmp;
  }


  // Not in cache ...
  DdNode        *tmp0, *tmp1, *tmp2;
  int           f_index, g_index;

  f_index = cuddIZ(dd, f->index);
  g_index = cuddIZ(dd, g->index);

  
  // Cases ...
  if (f_index > g_index)
    return(cuddZddUnionSA(dd, g, f));  
    

  if(f_index < g_index) {

    tmp1 = cuddZddNoSubA(dd, cuddT(f));
    if (tmp1 == NULL) {
      return(NULL);
      }
    cuddRef(tmp1);

    tmp0 = cuddZddUnionSA(dd, cuddE(f), g);
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);

    // Compute Diff
    tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
    if (tmp2 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddRef(tmp2);
    Cudd_RecursiveDerefZdd(dd, tmp1);

    tmp = cuddZddGetNode(dd, f_index, tmp2, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp2);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddRef(tmp);
    cuddDeref(tmp2); 
    cuddDeref(tmp0); 
    

    // Check 1 level only
    tmp1 = cuddZddNoSubAF(dd, tmp);
    if (tmp1 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp);
      return(NULL);
    }
    cuddRef(tmp1);
    Cudd_RecursiveDerefZdd(dd, tmp);
    tmp = tmp1;
    
  }
  else {
    tmp1 = cuddZddUnionSA(dd, cuddT(f), cuddT(g));
    if (tmp1 == NULL) {
      return(NULL);
    }
    cuddRef(tmp1);

    tmp0 = cuddZddUnionSA(dd, cuddE(f), cuddE(g));
    if (tmp0 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      return(NULL);
    }
    cuddRef(tmp0);
    
    tmp2 = cuddZddSubDiff(dd, tmp1, tmp0);
    if (tmp2 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp1);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddRef(tmp2);
    Cudd_RecursiveDerefZdd(dd, tmp1);

    tmp = cuddZddGetNode(dd, f->index, tmp2, tmp0);
    if (tmp == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp2);
      Cudd_RecursiveDerefZdd(dd, tmp0);
      return(NULL);
    }
    cuddRef(tmp);
    cuddDeref(tmp2);
    cuddDeref(tmp0);

    tmp1 = cuddZddNoSubAF(dd, tmp);
    if (tmp1 == NULL) {
      Cudd_RecursiveDerefZdd(dd, tmp);
      return(NULL);
    }
    cuddRef(tmp1);
    Cudd_RecursiveDerefZdd(dd, tmp);
    tmp = tmp1;
    
  }

  cuddCacheInsert2(dd, cuddZddUnionSA, f, g, tmp);
  cuddDeref(tmp);

  return tmp; 
  
}

